/* lsmbox -- list nr of total and read messages for one or several mailboxes
 *
 * Copyright (C) 2002 David Weinehall
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define _GNU_SOURCE
#include <pwd.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#undef NDEBUG
#include <assert.h>
#ifdef USE_CURSES
#include <curses.h>
#endif /* USE_CURSES */
#include "getopt.h"
#include <limits.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>

#define PRG_NAME		"lsmbox"
#define PRG_VERSION		"v1.4.2"

#define SPOOLFILE_PATH		"/var/mail"	/* username follows */
#define MAILBOX_FOLDER_PATH	"Mail"		/* $HOME preceeds */

#define SET_STR			"set"
#define UNSET_STR		"unset"
#define MARK_OLD_STR		"mark_old"
#define SPOOLFILE_STR		"spoolfile"
#define MAILBOX_FOLDER_STR	"folder"
#define MAILBOX_STR		"mailboxes"

#define FROM_STR		"From "
#define STATUS_STR		"Status: "
#define CONTENT_LENGTH_STR	"Content-Length: "

#define BUF_SIZE		1024

#if !defined __GNUC__ || __GNUC__ == 2 && __GNUC_MINOR__ < 96
#define __builtin_expect(__x, __expected_value) (__x)
#endif

#ifndef likely
#define likely(__x)       __builtin_expect((__x), 1)
#endif /* likely */

#ifndef unlikely
#define unlikely(__x)     __builtin_expect((__x), 0)
#endif /* unlikely */

#ifdef USE_CURSES
#define OPTSTR		"cnost"
#else
#define OPTSTR		"nost"
#endif /* USE_CURSES */

	struct option const options[] = {
#ifdef USE_CURSES
		{ "continuous", no_argument, 0, 'c' },
#endif /* USE_CURSES */
		{ "new-only", no_argument, 0, 'n' },
		{ "no-old", no_argument, 0, 'o' },
		{ "no-summary", no_argument, 0, 's' },
		{ "no-total", no_argument, 0, 't' },
		{ "help", no_argument, 0, 'h' },
		{ "version", no_argument, 0, 'V' },
		{ 0, 0, 0, 0 }
	};

typedef enum {
	E_OK,
	E_SYNTAX,
	E_UNKNOWN
} lsmbox_error_t;

char buf[BUF_SIZE];

char *homepath = NULL;
char *spoolfilepath = NULL;
char *mailfolderpath = NULL;
char **mailboxes = NULL;
int markold = 1;
int continuous = 0;
int nosummary = 0;
int newonly = 0;
int nototal = 0;
int noold = 0;
int quit = 0;

struct mailstat {
	char *fname;
	int exists;
	unsigned long total;
	unsigned long new;
	unsigned long old;
	off_t size;
	time_t mtime;
};

struct mailstat *mboxes = NULL;

/**
 * Show usage-information for the program
 *
 * @version v1.00
 */
void usage(void)
{
	fprintf(stdout,
		"Usage: %s [OPTION]... [MAILBOX]...\n"
		"List the number of total, old unread, and new messages\n"
		"for the specified mailbox(es).\n"
		"\n"
#ifdef USE_CURSES
		"  -c, --continuous  update statistics continuously\n"
#endif /* USE_CURSES */
		"  -n, --new-only    only display mailboxes with new messages\n"
		"  -o, --no-old      do not display number of old messages\n"
		"                      for each mailbox\n"
		"  -s, --no-summary  do not display a summary\n"
		"  -t, --no-total    do not display total number of messages\n"
		"                      for each mailbox\n"
		"      --help        display this help and exit\n"
		"      --version     output version information and exit\n"
		"\n"
		"Report bugs to <tao@kernel.org>\n", PRG_NAME);
}

/**
 * Show version-information for the program
 *
 * @version v1.00
 */
void version(void)
{
	fprintf(stdout,
		"%s %s\n"
		"Written by David Weinehall.\n"
		"\n"
		"Copyright (C) 2002 David Weinehall (tao@kernel.org)\n"
		"This is free software; see the source for copying "
		"conditions.  There is NO\n"
		"warranty; not even for MERCHANTABILITY or FITNESS FOR "
		"A PARTICULAR PURPOSE.\n",
		PRG_NAME, PRG_VERSION);
}

/**
 * Reset curses
 *
 * @version v1.00
 */
void reset_curses(void)
{
#ifdef USE_CURSES
	endwin();
#else
	;
#endif /* USE_CURSES */
}

/**
 * Setup curses
 *
 * @version v1.01
 *
 *  @return 1 on success, 0 on failure
 */
int init_curses(void)
{
#ifdef USE_CURSES
	initscr();
	assert(!atexit(reset_curses));

	if (cbreak() == ERR || noecho() == ERR || nonl() == ERR ||
	    curs_set(0) == ERR || nodelay(stdscr, 1) == ERR ||
	    intrflush(stdscr, 0) == ERR)
		return 0;

	assert(refresh() != ERR);
#endif /* USE_CURSES */

	return 1;
}

/**
 * Update screen
 *
 * @version v1.00
 */
void update_screen(void)
{
#ifdef USE_CURSES
	if (continuous)	{
		clrtobot();
		assert(refresh() != ERR);
		assert(move(0, 0) != ERR);
	}
#else
	;
#endif /* USE_CURSES */
}

/**
 * Enable bold text-attribute
 *
 * @version v1.00
 */
void bold_on(void)
{
#ifdef USE_CURSES
	if (continuous)
		attron(A_BOLD);
#else
	;
#endif /* USE_CURSES */
}

/**
 * Disable bold text-attribute
 *
 * @version v1.00
 */
void bold_off(void)
{
#ifdef USE_CURSES
	if (continuous)
		attroff(A_BOLD);
#else
	;
#endif /* USE_CURSES */
}

/**
 * Wrapper for getch()
 *
 * @version v1.00
 *
 *  @return the keypress or 0 if no key was pressed
 */
int curses_getch(void)
{
#ifdef USE_CURSES
	int c = 0;

	if (continuous)
		c = getch();

	return c == ERR ? 0 : c;
#else
	return 0;
#endif /* USE_CURSES */
}

/**
 * Output a formatted string
 *
 * @version v1.00
 *
 *   @param fmt The formatting string
 *   @param args The rest of the arguments
 */
void output_string(const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
#ifdef USE_CURSES
	if (continuous)
		vw_printw(stdscr, (char *)fmt, args);
	else
		vfprintf(stdout, fmt, args);
#else
	vfprintf(stdout, fmt, args);
#endif /* USE_CURSES */
	va_end(args);
}

/**
 * Wrapper for strdup that checks for NULL
 *
 * @version v1.00
 *
 *   @param str A pointer to the string to duplicate
 *
 *  @return NULL if str is NULL, the duplicated string otherwise
 */
char *strdup_null(const char *str)
{
	return str ? strdup(str) : NULL;
}

/**
 * Copy a string and return a pointer to its end
 *
 * @version v1.00
 *
 *   @param dst A pointer to the destination
 *   @param src A pointer to the source
 *
 *  @return A pointer to the end of the destination string
 */
char *internal_stpcpy(char *dst, const char *src)
{
	if (!dst || !src)
		return NULL;

	while ((*dst++ = *src++) != '\0')
		/* nothing */;

	return dst - 1;
}

/**
 * Join a variable amount of strings,
 * with an optional delimiter between each
 *
 * @version v1.00
 *
 *   @param delimiter
 *   @param src A pointer to the source
 *
 *  @return A newly allocated string with the result
 */
char *strjoin(char *delimiter, ...)
{
	va_list args;
	size_t slen;
	size_t len;
	char *str;
	char *tmp;
	char *ptr;

	if (delimiter == NULL)
		delimiter = "";

	slen = strlen(delimiter);

	va_start(args, delimiter);

	if ((tmp = va_arg(args, char *))) {
		len = strlen(tmp) + 1;

		while ((tmp = va_arg(args, char *)))
			len += slen + strlen(tmp);

		va_end(args);

		str = calloc(len, sizeof (char));

		va_start(args, delimiter);

		tmp = va_arg(args, char *);
		ptr = internal_stpcpy(str, tmp);

		while ((tmp = va_arg(args, char *))) {
			ptr = internal_stpcpy(ptr, delimiter);
			ptr = internal_stpcpy(ptr, tmp);
		}
	} else {
		str = strdup_null("");
	}

	va_end(args);

	return str;
}

/**
 * Split a string into an array, using the specified delimiter
 * to tell where to make the splits
 *
 * @version v1.00
 *
 *   @param str A pointer to the string to split
 *   @param delimiter A pointer to the delimiter
 *
 *  @return A NULL-terminated, newly allocated array of strings
 */
char **strsplit(const char *str, const char *delimiter)
{
	size_t asize = 1;
	size_t dlen;
	const char *tmp2;
	char **array = NULL;
	char *tmp;

	if (!str || !delimiter || !delimiter[0])
		return NULL;

	dlen = strlen(delimiter);

	tmp2 = str;

	while ((tmp = strstr(tmp2, delimiter))) {
		size_t len;
		char *new;

		len = tmp - tmp2;
		new = calloc(len + 1, sizeof (char));
		strncpy(new, tmp2, len);
		new[len] = '\0';

		asize++;
		array = realloc(array, sizeof (char *) * asize);
		array[asize - 2] = new;

		tmp2 = tmp + dlen;
	}

	/* add the remaining string */
	asize++;
	array = realloc(array, sizeof (char *) * asize);
	array[asize - 2] = strdup_null(tmp2);

	/* NULL terminate the array */
	array[asize - 1] = NULL;

	return array;
}

/**
 * Skip all leading whitespace in a string
 *
 * @version v1.00
 *
 *   @param str The string to process
 *
 *  @return A new pointer into the string with leading whitespace skipped
 */
char *skip_whitespace(char *str)
{
	for (; *str != 0; str++)
		if (!isspace((int)*str))
			break;
	return str;
}

/**
 * Skip past a configuration-option to its value
 *
 * @version v1.00
 *
 *   @param str The string to process
 *
 *  @return A new pointer to the value of the configuration-option
 */
char *skip_config_option(char *str)
{
	for (; *str != 0; str++)
		if (isspace((int)*str) || *str == '=')
			break;

	for (; *str != 0; str++)
		if (!isspace((int)*str) && *str != '=')
			break;

	return str;
}

/**
 * Get the user-name of the caller of the program
 *
 * @version v1.00
 *
 *  @return A pointer to the user-name
 */
const char *get_user_name(void)
{
	struct passwd *pw;

	assert((pw = getpwuid(getuid())));

	return pw->pw_name;
}

/**
 * Free a NULL-terminated array of strings, and the array itself
 *
 * @version v1.00
 *
 *   @param array The array to free
 */
void strfreev(char **array)
{
	int i;

	if (!array)
		return;

	for (i = 0; array[i]; i++)
		free(array[i]);

	free(array);
}

/**
 * Expand a path-name
 *
 * @version v1.00
 *
 *   @param str The path to expand
 *
 *  @return A pointer to the string with the new path;
 *          needs to be freed after use
 */
char *path_expand(const char *str)
{
	char *tmp = NULL, *tmp2 = NULL;
	struct passwd *pw;

	if (!strcmp(str, "!"))
		tmp = strdup_null(spoolfilepath);
	else if (strchr(str, '/'))
		tmp = strdup_null(str);
	else if (*str == '+' || *str == '=')
		tmp = strjoin("/", mailfolderpath, str + 1, NULL);
	else
		tmp = strjoin("/", mailfolderpath, str, NULL);

	if (*tmp == '~') {
		char *rest = NULL;

		if (*(tmp + 1) == '/') {
			assert((pw = getpwuid(getuid())));
		} else {
			char *name = strdup_null(tmp + 1);
			char *end;
			end = strchr(name, '/');
			*end = '\0';
			assert((pw = getpwnam(name)));
			free(name);
		}
		rest = strchr(tmp, '/') + 1;
		tmp2 = strjoin("/", pw->pw_dir, rest, NULL);
	} else {
		tmp2 = strdup_null(tmp);
	}

	free(tmp);

	return tmp2;
}

/**
 * Parse the configuration-file
 *
 * @version v1.07
 *
 *   @param fname The name of the configuration-file to parse
 *
 *  @return 1 on success, 0 on failure
 */
int parse_config_file(const char *fname)
{
	FILE *fp;
	char *str;
	char *mb = NULL;
	int line = 0;
	lsmbox_error_t error;

	if (!(fp = fopen(fname, "r"))) {
		if (errno == ENOENT)
			errno = 0;
		return !errno;
	}

	/* Poison */
	buf[BUF_SIZE - 1] = (char)0xff;

	while ((str = fgets(buf, BUF_SIZE, fp))) {
		/* We got a single line larger than 64kB */
		if (buf[BUF_SIZE - 1] != (char)0xff) {
			errno = EFBIG;
			return 0;
		}

		error = E_OK;

		line++;

		/* Ignore all comments */
		if (*str == '#')
			continue;

		str = skip_whitespace(str);

		/* Skip empty lines */
		if (!*str)
			continue;

		/* Remove the trailing newline */
		*(str + strlen(str) - 1) = '\0';

		if (!strncmp(str, SET_STR, strlen(SET_STR))) {
			str = skip_config_option(str);

			if (!strncmp(str, SPOOLFILE_STR,
				     strlen(SPOOLFILE_STR))) {
				str = skip_config_option(str);

				if (!str) {
					error = E_SYNTAX;
				} else {
					free(spoolfilepath);
					spoolfilepath = strdup_null(str);
				}
			} else if (!strncmp(str, MAILBOX_FOLDER_STR,
					    strlen(MAILBOX_FOLDER_STR))) {
				str = skip_config_option(str);

				if (!str) {
					error = E_SYNTAX;
				} else {
					free(mailfolderpath);
					mailfolderpath = strdup_null(str);
				}
			} else if (!strncmp(str, MARK_OLD_STR,
					    strlen(MARK_OLD_STR))) {
				markold = 1;
			} else {
				error = E_UNKNOWN;
			}
		} else if (!strncmp(str, UNSET_STR, strlen(UNSET_STR))) {
			str = skip_config_option(str);

			if (!strncmp(str, SPOOLFILE_STR,
				     strlen(SPOOLFILE_STR))) {
				free(spoolfilepath);
				spoolfilepath = NULL;
			} else if (!strncmp(str, MAILBOX_FOLDER_STR,
					    strlen(MAILBOX_FOLDER_STR))) {
				free(mailfolderpath);
				mailfolderpath = NULL;
			} else if (!strncmp(str, MARK_OLD_STR,
				strlen(MARK_OLD_STR))) {
				markold = 0;
			} else {
				error = E_UNKNOWN;
			}
		} else if (!strncmp(str, MAILBOX_STR, strlen(MAILBOX_STR))) {
			char *tmp;

			str = skip_config_option(str);

			if (!str) {
				error = E_SYNTAX;
			} else {
				if (mb)
					tmp = strjoin(" ", mb, str, NULL);
				else
					tmp = strdup_null(str);

				free(mb);
				mb = tmp;
			}
		} else {
			error = E_SYNTAX;
		}

		switch (error) {
		case E_UNKNOWN:
			fprintf(stderr,
				"Unknown option on line: %d inside: %s;"
				"ignoring.\n",
				line, fname);
			break;
		case E_SYNTAX:
			fprintf(stderr,
				"Syntax error on line: %d inside: %s;"
				"ignoring.\n",
				line, fname);
			break;
		case E_OK:
		default:
			break;
		}
	}

	if (mb)
		mailboxes = strsplit(mb, " ");

	free(mb);

	return !(fclose(fp) == EOF);
}

/**
 * Parse a mailbox in mbox/MMDF format
 *
 * @version v1.12
 *
 *   @param mbox A mailstat structure for the mailbox
 *
 *  @return 1 on success, 0 on failure
 */
int parse_mailbox(struct mailstat *mbox)
{

	FILE *fp;
	char *str;
	unsigned long clen = 0;
	int msg = 0;
	int old = 0;
	int mread = 1;
	int msgbody = 0;

	mbox->total = 0;
	mbox->new = 0;
	mbox->old = 0;

	if (!(fp = fopen(mbox->fname, "r")))
		return 0;

	/* Poison */
	buf[BUF_SIZE - 1] = (char)0xff;

	while (likely(str = fgets(buf, BUF_SIZE, fp))) {
		/* We got a single line larger than 64kB */
		if (unlikely(buf[BUF_SIZE - 1] != (char)0xff)) {
			errno = EFBIG;
			goto EXIT;
		}

		if (unlikely(!strncmp(str, FROM_STR,
				      strlen(FROM_STR)))) {
			msg = 1;
			mbox->total++;

			if (old)
				mbox->old++;
			else if (!mread)
				mbox->new++;

			old = 0;
			mread = 0;
			msgbody = 0;

			if (continuous && curses_getch() == 'q') {
				quit = 1;
				goto EXIT;
			}
		} else if (likely(msgbody)) {
			continue;
		} else if (unlikely(!strncmp(str, STATUS_STR,
					     strlen(STATUS_STR)))) {
			if (str[8] == 'R')
				mread = 1;
			else if (str[8] == 'O')
				old = 1;
		} else if (unlikely(!strncmp(str, CONTENT_LENGTH_STR,
				             strlen(CONTENT_LENGTH_STR)))) {
			clen = strtoul(str + strlen(CONTENT_LENGTH_STR),
				       NULL, 10);

		        if (clen == ULONG_MAX && errno == ERANGE) {
		                errno = ERANGE;
				goto EXIT;
			} else {
				if (fseek(fp, clen, SEEK_CUR))
					goto EXIT;
				clen = 0;
			}
		} else if (*str == '\n') {
			msgbody = 1;
		}
	}

	if (old)
		mbox->old++;
	else if (!mread)
		mbox->new++;

EXIT:
	return !(fclose(fp) == EOF);
}

/**
 * Show statistics for all mailboxes in the mboxes array
 *
 * @version v1.05
 *
 *  @return 0 on success, errno on failure
 */
int show_mailbox_stats(void)
{
	int i;
	unsigned long new;
	struct mailstat totals;
	int updated = 0;

	for (i = 0; mboxes[i].fname; i++) {
		struct stat st;

		if (stat(mboxes[i].fname, &st)) {
			if (errno == ENOENT) {
				mboxes[i].exists = 0;
				errno = 0;
				continue;
			} else {
				goto EXIT;
			}
		}

		mboxes[i].exists = 1;

		if (st.st_mtime <= mboxes[i].mtime &&
		    st.st_size == mboxes[i].size)
			continue;

		if ((!parse_mailbox(&mboxes[i]) && errno != ENOENT) || quit)
			goto EXIT;

		if (errno == ENOENT) {
			errno = 0;
			continue;
		}

		updated = 1;
	}

	if (!updated)
		goto EXIT;

	totals.total = 0;
	totals.new = 0;
	totals.old = 0;

	bold_on();
	output_string("mailbox:                                     "
		"%s%s:%s\n",
		nototal ? "" : "total:      ",
		(markold || noold) ? "new" : "unread",
		(markold && !noold) ? "        old:" : "");
	bold_off();

	for (i = 0; mboxes[i].fname; i++) {
		if (!mboxes[i].exists)
			continue;

		if (markold || noold)
			new = mboxes[i].new;
		else
			new = mboxes[i].new + mboxes[i].old;

		if (newonly && !new)
			continue;

		totals.total += mboxes[i].total;
		totals.new += new;
		totals.old += mboxes[i].old;

		output_string("%-43s  ", mboxes[i].fname);

		if (!nototal)
			output_string("%-10lu  ", mboxes[i].total);

		output_string("%-10lu", new);

		if (markold && !noold)
			output_string("  %-10lu", mboxes[i].old);

		output_string("\n");
	}

	if (!nosummary) {
		output_string("\nSummary:                                     ");

		if (!nototal)
			output_string("%-10lu  ", totals.total);

		if (markold || noold)
			new = totals.new;
		else
			new = totals.new + totals.old;

		output_string("%-10lu", new);

		if (markold && !noold)
			output_string("  %-10lu\n", totals.old);

		output_string("\n");
	}

	update_screen();

EXIT:

	return errno;
}

/**
 * The program's main-function
 *
 * @version v1.14
 *
 *   @param argc The number of arguments
 *   @param argv The arguments
 *
 *  @return 0 on success, errno on failure
 */
int main(int argc, char *argv[])
{
	int i;
	int optc;
	int opt_index;
	int start, end;
	extern int optind;

	char **list = NULL;
	char *userrcpath = NULL;

	errno = 0;

	while ((optc = getopt_long(argc, argv, OPTSTR,
				   options, &opt_index)) != -1) {
		switch (optc) {
		case 'c':
			continuous = 1;
			break;
		case 'n':
			newonly = 1;
			break;
		case 'o':
			noold = 1;
			break;
		case 's':
			nosummary = 1;
			break;
		case 't':
			nototal = 1;
			break;
		case 'h':
			usage();
			goto EXIT;
		case 'V':
			version();
			goto EXIT;
		default:
			usage();
			errno = EINVAL;
			goto EXIT;
		}
	}

	homepath = strdup_null(getenv("HOME"));

	if (homepath) {
		userrcpath = strjoin(NULL, homepath, "/.lsmboxrc", NULL);

		if (!parse_config_file(userrcpath))
			goto EXIT;
	}

	if (!spoolfilepath) {
		if (!(spoolfilepath = strdup_null(getenv("MAIL")))) {
			spoolfilepath = strdup_null(SPOOLFILE_PATH);
			spoolfilepath = strjoin("/", SPOOLFILE_PATH,
						  get_user_name(), NULL);
		}
	}

	if (!mailfolderpath) {
		if (homepath)
			mailfolderpath = strjoin("/", homepath,
						 MAILBOX_FOLDER_PATH, NULL);
		else
			mailfolderpath = strdup(".");
	}

	if ((argv[optind] == NULL || (argc < 2)) && !mailboxes) {
		usage();
		errno = EINVAL;
		goto EXIT;
	}

	if (argv[optind]) {
		start = optind;
		end = argc;
		list = argv;
	} else {
		start = 0;
		for (end = 0; mailboxes[end]; end++)
			;
		list = mailboxes;
	}

	if (continuous)
		if (!init_curses())
			goto EXIT;

	mboxes = calloc(end - start + 1, sizeof (struct mailstat));

	for (i = start; i < end; i++)
		mboxes[i - start].fname = path_expand(list[i]);

	do {
		show_mailbox_stats();
	} while (!errno && continuous && !quit);

EXIT:
	for (i = 0; mboxes && mboxes[i].fname; i++)
		free(mboxes[i].fname);

	free(mboxes);
	free(spoolfilepath);
	free(userrcpath);
	free(mailfolderpath);
	strfreev(mailboxes);

	return errno;
}
